<?php
/**
 * @file
 * Date/time API functions
 */

/**
 *  use pear date for advanced date handling functions and ability to process dates earlier than 1/1/1970
 *  also has improved handling for timezone and dst conversions
 */

define('DATE_FORMAT_DRUPAL', 3);
define('DATE_FORMAT_STRING_DRUPAL', "%Y%m%dT%H:%M:%S");

if (!include_once('Date.php')) drupal_set_message(t('The PEAR date files are not available at '.drupal_get_path('module', 'date').'/pear and date functions may not work correctly. Please contact the site administrator.'), 'error');

/**
 *  Date/Time Drop-Down Selector
 * 
 *  flexible date selector designed to be available to other modules as an api
 * 
 *  $params = an array of values, including
 *    label = the label for the date group, default is 'Date'
 *    value = the date/time to be processed, default is the current date and time
 *    timezone_in  = the timezone id of the date/time value to be processed, default is GMT
 *    timezone_out = the timezone id to be used when displaying the date, default is date site timezone, if set, or GMT
 *       valid timezone ids are standard timezone ids like US/Central, America/New_York, GMT
 *    format = the format of the date value, default is I
 *       U => unix timestamp
 *       I => iso 8601 (YYYY-MM-DDThh:mm:ss)
 *    am/pm = 0 to display 24 hour format, 1 to display 12 hour format with am/pm selector, default is 0
 *    weight = the weight of the date group, default is 0
 *    delta = a delta value for the group to accomodate multiple date fields, default is 0
 *    granularity = an array of date parts to be selected, like array('Y','M','D'), default is M, D, Y
 *      Y => year
 *      M => month
 *      D => day
 *      H => hour
 *      N => minute
 *      S => second
 *      T => timezone
 *    increment = increment minutes and seconds by what amount, default is 1
 *    opt_fields = an array of fields that need not be filled out, default is empty array
 *    blank_default = 1 to show an empty date field with blank values, 0 to fill with current date, default is 0
 *    required = 1 if the field must contain a valid date, default is 1
 *    description = text to be used as a description for the fieldset
 */
function date_selector($params) {
  
  theme_add_style(drupal_get_path('module', 'date') .'/date.css');
  
  // set the variables
  $label       = $params['label'] ? $params['label'] : t('Date');
  $delta       = isset($params['delta']) ? $params['delta'] : 0;
  $title       =  ($delta > 0) ? (' '.$delta) : '';
  $granularity = is_array($params['granularity']) ? $params['granularity'] : array('M', 'D', 'Y');
  $increment   = isset($params['increment']) ? $params['increment'] : 1;
  $required    = isset($params['required']) ? $params['required'] : 1;
  $format      = isset($params['format']) ? $params['format'] : 'I';
  $weight      = isset($params['weight']) ? $params['weight'] : 0;
  $show_ampm   = isset($params['ampm']) ? $params['ampm'] : 0;
  $required    = isset($params['required']) ? $params['required'] : 1;
  $opt_fields  = is_array($params['opt_fields']) ? $params['opt_fields'] : array();
  $blank_default = isset($params['blank_default']) ? $params['blank_default'] : 0;
  $timezone_in   = isset($params['timezone_in']) ? $params['timezone_in'] : 'GMT';
  $timezone_out  = isset($params['timezone_out']) ? $params['timezone_out'] : date_get_site_timezone();
  $description   = $params['description'];

  if ($ampm) {
    $hour_format = t('%I');
    for ($i = 0; $i <= 12; $i++) $hours[$i] = $i < 10 ? "0$i" : $i;
  } 
  else {
    $hour_format = t('%H');
    for ($i = 0; $i <= 23; $i++) $hours[$i] = $i < 10 ? "0$i" : $i;
  }
    
  // create a pear date object with the desired date
  $date = new Date();
  switch ($format) {
  case ('U'):
    $type = DATE_FORMAT_UNIXTIME;
    if ($params['value']) $date->setDate($params['value'], $type);
    break;
  default:
    $type = DATE_FORMAT_DRUPAL;
    if ($params['value'] > '') $date->setDate($params['value'], $type); 
    break;
  }

  // if setting new date to current time, eliminate minutes and seconds
  if (!$blank_default && $params['value'] == '') {
    $date->setDate(time(), $type);
  }

  // set the timezone to the input value, then convert the date to the timezone to be displayed
  date_convert_timezone($date, $timezone_in, $timezone_out);
    
  if (!$blank_default || $params['value'] > '') {
    $year = $date->format('%Y');
    $month = intval($date->month);
    $day = $date->format('%d');
    $hour = $date->format($hour_format);
    $minute = (string)$date->format('%M');
    $second = (string)$date->format('%S');
    $ampm = $date->format('%p');
  } 

  // create the form values
  $form['#type']   = 'fieldset';
  $form['#title']  = t($label) . $title;
  $form['#weight'] = intval($weight);
  $form['#prefix'] = '<div class="container-inline-date">';
  $form['#suffix'] = '</div>';
  
  if (in_array('M',$granularity)) {
    $months = array(1 => t('January'), t('February'), t('March'), t('April'), t('May'), t('June'), t('July'), t('August'), t('September'), t('October'), t('November'), t('December'));
    if (!$required || in_array('month', $opt_fields) || $blank_default) array_unshift($months, '');
    $form['month'] = array(
    '#type' => 'select',
    '#default_value' => $month,
    '#options' => $months,
    '#title' => t('month'),
    '#required' => ($required && !in_array('month', $opt_fields)) ? $required : 0,
    );
  }
  if (in_array('D',$granularity)) {
    $form['day'] = array(
    '#type' => 'textfield',
    '#default_value' => $day,
    '#maxlength' => 2,
    '#size' => 2,
    '#title' => t('day'),
    '#required' => ($required && !in_array('day', $opt_fields)) ? $required : 0,
    );
  }
  if (in_array('Y',$granularity)) {
    $form['year'] = array(
    '#type' => 'textfield',
    '#default_value' => $year,
    '#maxlength' => 4,
    '#size' => 4,
    '#title' => t('year'),
    '#required' => ($required && !in_array('year', $opt_fields)) ? $required : 0,
    );
  }
  $form['space1'] = array(
    '#type' => 'markup',
    '#value' => '<div class="form-item">&nbsp;&nbsp;</div>',
    );
  if (in_array('H',$granularity)) {
    $form['hour'] = array(
    '#type' => 'select',
    '#default_value' => $hour,
    '#options' => $hours,
    '#required' => ($required && !in_array('hour', $opt_fields)) ? $required : 0,
    '#title' => t('hour'),
    );
  }
  if (in_array('N',$granularity)) {
    for ($i = 0; $i <= 59; $i += $increment) $minutes[$i] = $i < 10 ? "0$i" : $i;
    $form['minute'] = array(
    '#type' => 'select',
    '#default_value' => $minute,
    '#options' => $minutes,
    '#required' => ($required && !in_array('minute', $opt_fields)) ? $required : 0,
    '#title' => t('min'),
    );
  }
  if (in_array('S',$granularity)) {
    for ($i = 0; $i <= 59; $i += $increment) $seconds[$i] = $i < 10 ? "0$i" : $i;
    $form['second'] = array(
    '#type' => 'select',
    '#default_value' => $second,
    '#options' => $seconds,
    '#required' => ($required && !in_array('second', $opt_fields)) ? $required : 0,
    '#title' => t('sec'),
    );
  }
  if ($show_ampm) {
    $options = array('am' => t('am'), 'pm' => t('pm'));
    if (!$required || in_array('hour', $opt_fields) || $blank_default) array_unshift($options, '');
    $form['space2'] = array(
      '#type' => 'markup',
      '#value' => '<div class="form-item">&nbsp;</div>',
      );
    $form['ampm'] = array(
      '#type' => 'select',
      '#default_value' => $ampm,
      '#options' => $options,
      '#required' => ($required && !in_array('hour', $opt_fields)) ? $required : 0,
      '#title' => t('am/pm'),
      );
  }
  if (in_array('T',$granularity)) {
    $form['timezone'] = array(
      '#type' => 'select',
      '#default_value' => $timezone_out,
      '#options' => date_timezone_options(),
      '#required' => ($required && !in_array('timezone', $opt_fields)) ? $required : 0,
      '#title' => t('timezone'),
      );
  } else {
    $form['timezone'] = array(
      '#type' => 'hidden',
      '#value' => $timezone_out,
      );
  }
  $form[] = array(
    '#type' => 'markup',
    '#value' => '<div class="form-item"><div class="description">'.$description.'</div></div>',
    );
  return $form;
}

/**
 *  Function to create database value from the value submitted by the above selector
 *   $out is array of values to be stripped from the date (H, M, S, etc)
 */
function date_selector_make_date($array, $type = 'I', $timezone_out = 'GMT') {
  
  switch ($type) {
  case ('U'):
    $format = DATE_FORMAT_UNIXTIME;
    break;
  default:
    $format = DATE_FORMAT_DRUPAL;
  }
  // try to construct a date from the submitted values
  $date = new Date();
  $date->setDate(date_array2iso($array));
  
  // if it cannot make a valid date, pear date gives the date a year of 0 and an hour of 0
  if ($date->year == 0 and $date->hour == 0) {
    return false;
  
  } else {
    // set the timezone to the input value, then convert the date to the timezone to be displayed
    if ($array['timezone'] > '') {
      date_convert_timezone($date, $array['timezone'], $timezone_out);
    }
    $out = $date->getDate($format);
    return $out;
  }
}

/**
 *  Validation function for above selector
 *  $params = an array of values include
 *    year_min = minimum year allowed, default is 1850
 *    year_max = maximum year allowed, default is 2050
 *    required = is a valid date required, default is true
 *    opt_fields = an array of fields that need not be filled out, default is empty array
 *    eliminate = an array of fields that should be eliminated from the date, default is empty array
 */
function date_selector_validate($value, $fieldname, $params = array()) {
  
  $type = $params['type'] > '' ? $params['type'] : 'I';
  $year_min = isset($params['year_min']) ? intval($params['year_min']) : 1850;
  $year_max = isset($params['year_max']) ? intval($params['year_max']) : 2050;
  $required = isset($params['required']) ? $params['required'] : 1;
  $opt_fields = is_array($params['opt_fields']) ? $params['opt_fields'] : array();
  
  // check for invalid numbers in fields that were submitted using textfields
  if ($required && !in_array($value['year'], $opt_fields)) {
    if ($value['year'] < $year_min || $value['year'] > $year_max || !is_numeric($value['year'])) form_set_error($fieldname, t('year must be a number between 1900 and 2050.'));
  }
  if ($required && !in_array($value['day'], $opt_fields)) {
    if ($value['day'] < 0 || $value['day'] > 31 || !is_numeric($value['day'])) form_set_error($fieldname, t('day must be a numer between 1 and 31.'));
  }
  // try to construct a date from the submitted values
  if ($required && !date_selector_make_date($value, $type)) {
    form_set_error($fieldname, t('Date cannot be constructed using values %s.', array('%s' => implode(', ',$value))));
    return false;
  } else {
    return true;
  }
}

/**
 * Date conversion functions
 *
 * content date field is stored in database in ISO 8601 format (YYYY-MM-DDThh:mm:ss) 
 * a variety of ways to convert values, not sure yet which will be needed
 * 
 */
function date_array2iso($array) {
  return  sprintf("%04d",intval($array['year'])) . 
    '-'. sprintf("%02d",intval($array['month'])) .  
    '-'. sprintf("%02d",intval($array['day'])) . 
    'T'. sprintf("%02d",intval($array['hour'])) . 
    ':'. sprintf("%02d",intval($array['minute'])) . 
    ':'. sprintf("%02d",intval($array['second']));
}

function date_iso2array($iso) {
  return array(
    'year' => substr($iso, 0, 4),
    'month' => substr($iso, 5, 2),
    'day' => substr($iso, 8, 2),
    'hour' => substr($iso, 11, 2),
    'minute' => substr($iso, 14, 2),
    'second' => substr($iso, 17,2),
    );
}

function date_unix2array($unix) {
  return array(
    'year' => format_date($unix, 'custom', 'Y'),
    'month' => format_date($unix, 'custom', 'm'),
    'day' => format_date($unix, 'custom', 'd'),
    'hour' => format_date($unix, 'custom', 'H'),
    'minute' => format_date($unix, 'custom', 'i'),
    'second' => format_date($unix, 'custom', 's'),
    );
}

function date_unix2iso($unix) {
  return date('Y-m-d\TH:i:s', $unix);
}

function date_iso2unix($iso) {
  return mktime(substr($iso, 11, 2), 
    substr($iso, 14, 2), 
    substr($iso, 17, 2), 
    substr($iso, 5, 2), 
    substr($iso, 8, 2), 
    substr($iso, 0, 4));
}

/**
 *  Use stringtotime function to create an iso date out of text
 */
function date_text2iso($text) {
  if ($str = @date('Y-m-d\TH:i:s', strtotime($text))) {
    echo $text.' translated to '.$str,'<br>';
    $date = new Date($str);
    return $date->format(DATE_FORMAT_STRING_DRUPAL);
  } else {
    return false;
  }
}

/**
 *  Validation functions
 *  Regex from original date module, may not be needed anymore, saving it just in case
 */
function date_preg($item) {
  if (preg_match('/[0-9]{8}T[0-9]{2}:[0-9]{2}:[0-9]{2}/', $item['value'])) {
    form_set_error($field['field_name'] .']['. $delta .'][value', t('%name must be entered in ISO 8601 format (YYYYMMDDThh:mm:ss).', array('%name' => t($field['widget']['label']))));
  }
}

/**
  *  Pear Date format, similar to strftime()
  *
  *  Formats the date in the given format, much like
  *  strftime().  Most strftime() options are supported.<br><br>
  *
  *  formatting options:
  *
  *  %a    abbreviated weekday name (Sun, Mon, Tue) 
  *  %A    full weekday name (Sunday, Monday, Tuesday) 
  *  %b    abbreviated month name (Jan, Feb, Mar) 
  *  %B    full month name (January, February, March) 
  *  %C    century number (the year divided by 100 and truncated to an integer, range 00 to 99) 
  *  %d    day of month (range 00 to 31) 
  *  %D    same as "%m/%d/%y" 
  *  %e    day of month, single digit (range 0 to 31) 
  *  %E    number of days since unspecified epoch (integer, Date_Calc::dateToDays()) 
  *  %H    hour as decimal number (00 to 23) 
  *  %I    hour as decimal number on 12-hour clock (01 to 12) 
  *  %j    day of year (range 001 to 366) 
  *  %m    month as decimal number (range 01 to 12) 
  *  %M    minute as a decimal number (00 to 59) 
  *  %n    newline character (\n) 
  *  %O    dst-corrected timezone offset expressed as "+/-HH:MM" 
  *  %o    raw timezone offset expressed as "+/-HH:MM" 
  *  %p    either 'am' or 'pm' depending on the time 
  *  %P    either 'AM' or 'PM' depending on the time 
  *  %r    time in am/pm notation, same as "%I:%M:%S %p" 
  *  %R    time in 24-hour notation, same as "%H:%M" 
  *  %s    seconds including the decimal representation smaller than one second 
  *  %S    seconds as a decimal number (00 to 59) 
  *  %t    tab character (\t) 
  *  %T    current time, same as "%H:%M:%S" 
  *  %w    weekday as decimal (0 = Sunday) 
  *  %U    week number of current year, first sunday as first week 
  *  %y    year as decimal (range 00 to 99) 
  *  %Y    year as decimal including century (range 0000 to 9999) 
  *  %%    literal '%' 
  */

/**
 * Format the view
 */
function date_iso2view($value, $format_string, $timezone_out = 'GMT', $timezone_in = 'GMT') {
  // process in output buffer to hide windows error messages for dates prior to 1970
  ob_start();
  $date = new Date();
  $date->setDate($value, DATE_FORMAT_DRUPAL);
  // set the timezone to the input value, then convert the date to the timezone to be displayed
  date_convert_timezone($date, $timezone_in, $timezone_out);
  $output = $date->format($format_string);
  $msg = ob_get_clean();
  return $output;
}

function date_unix2view($value, $format_string, $timezone_out = 'GMT', $timezone_in = 'GMT') {
  // process in output buffer to hide windows error messages for dates prior to 1970
  ob_start();
  $date = new Date();
  $date->setDate($value, DATE_FORMAT_UNIXTIME);
  // set the timezone to the input value, then convert the date to the timezone to be displayed
  date_convert_timezone($date, $timezone_in, $timezone_out);
  $msg = ob_get_clean();
  return $date->format($format_string);
}

/**
 *  Options for date view formats
 */
function date_date_options() {
  return array(
    ' ', '%Y', '%b', '%B', '%Y %b', '%b %Y', '%Y %B',
    '%B %Y', '%Y-%m-%d', '%m/%d/%Y', '%d/%m/%Y', 
    '%Y/%m/%d', '%b %e %Y', '%e %b %Y',
    '%Y %b %e', '%a, %Y-%m-%d', '%a, %m/%d/%Y', 
    '%a, %d/%m/%Y', '%a, %Y/%m/%d', '%B %e, %Y', 
    '%e %B, %Y', '%Y, %B %e', '%a, %m/%d/%Y', 
    '%a, %d/%m/%Y', '%a, %Y/%m/%d', '%B %e, %Y', 
    '%e %B %Y', '%Y, %B %e', '%A, %B %e, %Y', 
    '%A, %e %B, %Y', '%A, %Y,  %B %e',
    '%A, %B %e, %Y', '%A, %e %B %Y',
    '%A, %Y,  %B %e',
    );
}

/**
 * Options for time view formats
 */
function date_time_options() {
  return array(
    ' ',
    '%H:%M',
    '%H:%M %O',
    '%H:%M %Z',
    '%I:%M %p',
    '%I:%M %p %O',
    '%I:%M %p %Z',
    '%I:%M %P',
    '%I:%M %P %O',
    '%I:%M %P %Z',
    '%O', '%Z',
    );
}
/**
 *  Create visual examples of date format options in widget settings for each time format
 * 
 *  Use a date that will best illustrate formatting options
 *    - single digit months and days to see leading zeros
 *    - a month number different than the day number to see month and day positions
 *    - an afternoon time to illustrate 24 hour vs am/pm formats
 *  Selected date for illustration: August 1 YYYY 2:05:10 PM
 * 
 */
function date_view_options($options) {
  $date = new Date();
  $date->setDate(date('Y',time()).'0801T140510', DATE_FORMAT_DRUPAL);
  foreach ($options as $opt) {
    $view[$opt] = $date->format($opt);
  }
  return $view;
}

/**
 *  Time zone option list
 *  returns array of timezone ids (i.e. US/Eastern, US/Central)
 */
function date_timezone_options($limit = array()) {
  foreach ($GLOBALS['_DATE_TIMEZONE_DATA'] as $key => $val) {
    if (sizeof($limit) == 0 || (sizeof($limit) > 0 && in_array($key, $limit))) {
      $options[$key] = $key;
    }
  }
  asort($options);
  return $options;
}

/**
 *  Time zone value list
 *  returns array with the id as the key and the offset as the value
 *  offsets adjusted to match Drupal offsets (divide pear date timezone offsets by 1000)
 */
function date_timezone_vals($val_limit = array(), $key_limit = array()) {
  foreach ($GLOBALS['_DATE_TIMEZONE_DATA'] as $key => $val) {
    if ((sizeof($key_limit) == 0 && sizeof($val_limit) ==0 || sizeof($key_limit) > 0 && in_array($key, $key_limit)) || sizeof($val_limit) > 0 && in_array(intval($val['offset'] / 1000), $val_limit)) {
      $values[$key] = intval($val['offset'] / 1000);
    }
  }
  asort($values);
  return $values;
}

/**
 *  Get an array of timezones with an offset of $offset
 *  provides a way of reducing the size of the timezone list displayed
 */
function date_timezones_in_offset($offset) {
  $array = date_timezone_vals(array($offset));
  return array_keys($array);
}

/**
 *  Set system variable for site default timezone
 *  Needed because current system variable tracks offset rather than timezone
 */
function date_set_site_timezone($val) {
  if (!in_array($val, array_keys($GLOBALS['_DATE_TIMEZONE_DATA']))) $val = 'GMT';
  variable_set('date_field_timezone', $val);
}

/**
 *  Get system variable setting for site default timezone
 */
function date_get_site_timezone() {
  return variable_get('date_field_timezone', 'GMT');
}

/**
 *  Timezone handling options
 */
function date_timezone_handling_options() {

  // can't offer option to view in user's timzone until there is a system-wide variable available 
  // to store and retrieve the user's timezone instead of, or in addition to, their gmt offset
  // the gmt offset provides insufficent information to do proper timezone conversions
  
  return array(
    0 => 'GMT',
    1 => t('Site\'s time zone'),
    2 => t('Date\'s time zone'),
    //3 => t('User\'s time zone'), 
  );
}

/**
 *  Function to figure out which timezone applies to a date and select it
 */
function date_get_date_timezone($node_field, $field) {

  $handling_opts = date_timezone_handling_options();
  switch ($handling_opts[$field['widget']['tz_handling']]) {
    case ("Site's time zone"):
      $timezone_out = date_get_site_timezone();
      break;
    case ("Date's time zone"):
      $timezone_out = $node_field['timezone'] > '' ? $node_field['timezone'] : date_get_site_timezone();
      break;
    default:
      $timezone_out = 'GMT';
  }
  return $timezone_out;
}

/**
 *  Timezone conversion function
 *  uses pear date timezone conversion which can fail in some systems
 *  use output buffer to surpress windows error messages if year is less than 1970
 *  will fail silently if unable to do daylight savings time conversion
 *  worst case result should be 1 hour error
 */
function date_convert_timezone(&$date, $timezone_in, $timezone_out) {
  
  if ($timezone_in == 'GMT' && ($timezone_in == $timezone_out)) return;
  ob_start();
  $date->setTZbyID($timezone_in);
  $date->convertTZbyID($timezone_out);
  ob_get_clean();
}

/**
 *  Use Date_Calc to calculate the time since or until another date
 */
function date_days_to_now($date1) {
  if ($date1 == '') return false;
  // process in output buffer to hide windows error messages for dates prior to 1970
  ob_start();
  $date = new Date();
  $date->setDate(Date_Calc::dateNow());
  $day2   = $date->day;
  $month2 = $date->month;
  $year2  = $date->year;
  $date->setDate($date1);
  $future = $date->isFuture($date);
  $day1   = $date->day;
  $month1 = $date->month;
  $year1  = $date->year;
  $years  = abs($year1 - $year2);
  if ($years >= 1 and $future) {
    $days = Date_Calc::dateDiff($day1, $month1, intval($year2), $day2, $month2, $year2);
  } elseif ($years >= 1 && !$future) {
    $days = Date_Calc::dateDiff($day1, $month1, intval($year2), $day2, $month2, $year2);
  } else {
    $days = Date_Calc::dateDiff($day1, $month1, $year1, $day2, $month2, $year2);
  }
  $text   = ($years > 0) ? ($years . ($years > 1 ? t(' years ') : t(' year '))) : '';
  $text  .= $days > 0 ? ($days . ($days > 1 ? t(' days ') : t(' day '))) : '';
  $msg = ob_get_clean();
  if ($days + $years < 1) {
    return '('. t('now') . ')';
  } elseif ($future) {
    return ' ('. $text . t('from now'). ')';
  } else {
    return ' ('. $text . t('ago') . ')';
  }
}


